/* cbootloader.c           letzte Aenderung: 6.10.2012
Einfacher Bootloader fuer Atmega2560

Anwendung:
> Programm (z.B. Hexprog) am Computer starten
> Reset am Atmega2560 druecken
Abwarten bis Uebertragung fertig.

Ablauf des Datenaustauschs:
Bei einem Reset am Atmega2560 sendet der Bootloader ein Startzeichen ('A') und
wartet eine gewisse Zeit (35ms) auf Antwort.
Hexprog antwortet dann mit einem Schluesselwort ("StartBootloader")
Bootloader sendet dann "ok", oder bei falschem Schluesselwort "error".
Wenn keine Antwort kommt oder falsches Schluesselwort wird das Anwenderprogramm
bei 0x0000 gestartet.
Sonst beginnt die Bootloader-Routine
Hexprog sendet dann ein Kommando: "check", "readeprom", "readflash", "writeeprom" oder "writeflash".

Bei den read-Kommandos wird dann vom Bootloader eine Zeile im hex-Format gesendet,
ein "ok" von Hexprog abgewartet, naechste Zeile... bis ganzer Speicher ausgelesen.
Dann sendet der Bootloader ein "ok" um das Ende der Uebertragung anzuzeigen.

Bei den write-Kommandos sendet Hexprog eine Zeile im hex-Format, wartet "ok" vom
Bootloader ab, naechste Zeile... bis alle Daten uebertragen.
Dann sendet Hexprog "ok" um Ende der Uebertragung anzuzeigen.

Bei "check" wird vom Bootloader als Antwort jeweils die empfangene Zeile wieder zurueckgeschickt.
Wenn "reset" kommt, dann wird das Anwenderprogramm bei 0x0000 gestartet.


Autor: Rolf Pfister
Copyright: Opensource Freeware
*/
#define VERSION "0.1"
/*
29.9.2012    Erstellung

*/

#define ATMEGA2560
#define F_CPU 16000000UL
//#define BAUD  500000UL   //Baudrate
//#define BAUD 38400UL   //falls hohe Baudrate nicht geht
#define BAUD 9600UL   //falls hohe Baudrate nicht geht
//#define WARTEZEIT1 35 //Anzahl MilliSekunden zum auf erstes Zeichen warten
#define WARTEZEIT1 2000 //Anzahl MilliSekunden zum auf erstes Zeichen warten
#define WARTEZEIT2 5000 //Anzahl MilliSekunden wenn spaeter auf eine Antwort gewartet wird

#define TESTAUFLCD  //Testausdrucke auf dem LCD beim MK3
#define TESTLEDS    //Tests auf den LEDs am PORTL
#define DEBUG       //Zusaetzliche Tests zur Fehlersuche

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/boot.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <string.h>
#include "uart.c"

#ifdef TESTAUFLCD
#define LCD_MK3
#include "lcd_treiber.c"
#endif

#define MAXZ 80 //Maximale Zeilenlaenge
#define ERR_TIMEOUT 1

int8_t uart_zeile_empfangen(char *zeile,long timeout)
{
 char c;
 uint8_t i=0;
#if(F_CPU==16000000UL)
 timeout<<=10; //Bei 16MHz mit ca. 1000 multiplizieren
#else
 timeout<<=9; //Bei 8MHz weniger stark multiplizieren
#endif
 while(timeout>0)
  {if(uart_getcheck())
    {
     //c=uart_getc();
     c=UDR3;
     if(c=='\n' || i==MAXZ-1) {zeile[i]=0; return 0;}
     else if(c>=' ') zeile[i++]=c;
     else if(c=='\t') zeile[i++]=' '; //Tabulator durch Blank ersetzen, andere Steuerzeichen (z.B. '\r') ignorieren
    }
   else  --timeout;
  }
 zeile[i]=0;
 return ERR_TIMEOUT;
}

void uart_off()
{
 UCSR3B = 0;           //Auf Defaultwerte nach Datenblatt zuruecksetzen
 UCSR3C = (3<<UCSZ30); //Auf Defaultwerte nach Datenblatt zuruecksetzen
 UBRR3 = 0;            //Auf Defaultwerte nach Datenblatt zuruecksetzen
}

char hexdigit(uchar c)
{
 if(c<0x0A) return c+'0';
 return c-0x0A+'A';
}

int main()
{
 char zeile[MAXZ],antwort[MAXZ];
 int8_t err;
 uint8_t temp, n;
 void (*start)( void ) = 0x0000;  /* Funktionspointer auf 0x0000 */

#ifdef TESTAUFLCD
 //char text[MAXZ+20];
 uint8_t nzeilen=0;
#define maxzeilen 5
 glcd_init();
 glcd_clear();
 glcd_licht(true);
#endif

 /* Interrupt Vektoren verbiegen: */
 //char sregtemp = SREG; //nicht noetig weil Interrupt ja noch nicht eingeschaltet
 //cli();
 temp = MCUCR;
 MCUCR = temp | (1<<IVCE);
 MCUCR = temp | (1<<IVSEL);
 //SREG = sregtemp;

#ifdef TESTLEDS
 DDRL = 0xFF;
 PORTL = 0x05; //Beim Starten ein paar LEDs einschalten
#endif

 uart_init(); //Serielle Schnittstelle initialisieren
 sei();       //Interrupts einschalten

#ifdef TESTAUFLCD
 glcd_print("cbootloader Vers ");
 glcd_print(VERSION);
 glcd_print("\n");
#endif

#ifdef TESTLEDS
 PORTL = 0; //test: LEDs wieder aus
#endif

 /* Startbyte senden und auf Antwort mit Schluesselwort warten: */
#ifdef TESTAUFLCD
 glcd_print("gesendet:\"A\\n\"\n");
#endif
 //uart_putc('A');
 uart_write("A\n");
 err=uart_zeile_empfangen(zeile,WARTEZEIT1);
 if(err!=0 || strcmp(zeile,"StartBootloader")!=0)
  {
   antwort[0]=0;
   if(err==ERR_TIMEOUT)
        strcpy(antwort,"err-timeout1\n");
   else strcpy(antwort,"err-StartBootloader\n");
#ifdef TESTAUFLCD
   glcd_print("empfangen:'"); glcd_print(zeile); glcd_print("'\n"); _delay_ms(2000);
#endif
#ifdef DEBUG
   if(antwort[0]!=0) uart_write(antwort);
#endif
  }
 else //Bootloader-Routine nur machen wenn korrekt geantwortet wurde
  {
   uint8_t modus=0;
#define MODUS_CHECK 1
#define MODUS_WRITEFLASH 2
#define MODUS_WRITEEPROM 3
#define MODUS_READFLASH 4
#define MODUS_READEPROM 5
   uint16_t adresse=0; //16-Bit-Adresse
#ifdef TESTAUFLCD
   glcd_print("Empfangen:'"); glcd_print(zeile); glcd_print("'\n"); _delay_ms(2000);
#endif
   uart_write("ok\n");

   while(1) //Hauptschlaufe, wird nur mit break verlassen
    {
     err=uart_zeile_empfangen(zeile,WARTEZEIT2);
     if(err==ERR_TIMEOUT)
      {
#ifdef TESTLEDS
       PORTL = 3; _delay_ms(100); PORTL = 0; //test: 2 LEDs kurz blinken lassen
#endif
#ifdef TESTAUFLCD
       if(++nzeilen>maxzeilen)  {glcd_clear(); nzeilen=0;}
       glcd_print("TIMEOUT\n");
       uart_write("err-timeout2\n");
#endif
      }
     else //andernfalls Reaktion auf die empfangene Zeile 
      {
#ifdef TESTLEDS
       PORTL = 1; _delay_ms(100); PORTL = 0; //test: eine LED kurz blinken lassen
#endif
#ifdef TESTAUFLCD
       if(++nzeilen>maxzeilen)  {glcd_clear(); nzeilen=0;}
       glcd_print(zeile); //Empfangene Zeile auf dem LCD anzeigen
       glcd_print("\n");
#endif
       if(strcmp(zeile,"reset")==0) break; //bei "reset" Hauptschlaufe verlassen und Anwenderprogramm starten
       antwort[0]=0; //noch keine Antwort senden
       if(modus==0) //Anfangs auf Kommandos zum Modus setzen warten
	{if(strcmp(zeile,"writeflash")==0) {modus=MODUS_WRITEFLASH; strcpy(antwort,"ok\n");}
	 else if(strcmp(zeile,"writeeprom")==0) {modus=MODUS_WRITEEPROM; strcpy(antwort,"ok\n");}
	 else if(strcmp(zeile,"readflash")==0) modus=MODUS_READFLASH;
	 else if(strcmp(zeile,"readeprom")==0) modus=MODUS_READEPROM;
	 else if(strcmp(zeile,"check")==0) {modus=MODUS_CHECK; strcpy(antwort,"check\n");}
	 else strcpy(antwort,"err-command\n");
	}
       else if(modus==MODUS_CHECK)
	{
	 strcpy(antwort,zeile); antwort[n=strlen(zeile)]='\n'; antwort[n+1]=0;
	}
       else if(modus==MODUS_WRITEFLASH)
	{
	 if(strcmp(zeile,"ok")==0) break; //Ende der Uebertragung --> Hauptschlaufe verlassen
	 else if(zeile[0]!=':') strcpy(antwort,"err-colon\n");
	 else
	  {
#ifdef TESTAUFLCD
	   glcd_print("wuerde jetzt ins Flash gebrannt\n");
#endif
	   strcpy(antwort,"ok\n");
	  }
	}
       else if(modus==MODUS_WRITEEPROM)
	{
	 if(strcmp(zeile,"ok")==0) break; //Ende der Uebertragung --> Hauptschlaufe verlassen
	 else if(zeile[0]!=':') uart_write("err-colon\n");
	 else
	  {
#ifdef TESTAUFLCD
	   glcd_print("wuerde jetzt ins EEPROM gebrannt\n");
#endif
	   strcpy(antwort,"ok\n");
	  }
	}
       if(modus==MODUS_READFLASH)  //kein else weil sofort aufs Kommando "readflash" die erste Zeile gesendet werden soll
	{
#ifdef TESTAUFLCD
	 glcd_print("wuerde jetzt das Flash auslesen\n");
#endif
	 static int8_t testflag=0; //test
	 if(testflag==0) //test
	  {
	   strcpy(antwort,":100000000C9472000C947E000C947E000C947E0084\n"); //test
	   testflag=1;
	  }
	 if(antwort[0]==0) strcpy(antwort,"ok\n"); //wenn fertig ausgelesen, dann ok senden
	}
       else if(modus==MODUS_READEPROM)
	{
	 int8_t i;
	 if(strcmp(zeile,"ok")==0) //Wenn letzte 16 Zeichen erfolgreich empfangen wurden
	   adresse += 16;          //dann Adresse erhoehen. (Sonst Daten von selber Adresse nochmals senden)
	 if(adresse==0x1000) {uart_write("ok\n"); break;} //wenn fertig: "ok" senden und Hauptschlaufe verlassen
	 antwort[0]=':'; //die Hex-Daten in die Antwort-Zeile schreiben
	 for(i=0;i<16;i++)
	  {
	   uchar c = eeprom_read_byte((uint8_t*)(adresse+i));
	   antwort[1+2*i] = hexdigit(c>>4);
	   antwort[2+2*i] = hexdigit(c&0x0F);
	  }
	 antwort[1+2*16] = '\n';
	 antwort[2+2*16] = 0;
	}
#ifdef DEBUG
       if(antwort[0]==0) //wenn alles korrekt programmiert sollte das nicht passieren
	{strcpy(antwort,"err-unknown\n"); //test
#ifdef TESTLEDS
	 PORTL = 7; _delay_ms(100); PORTL = 0; //test: 3 LEDs kurz blinken lassen
#endif
#ifdef TESTAUFLCD
       if(++nzeilen>maxzeilen)  {glcd_clear(); nzeilen=0;}
       glcd_print(antwort);
#endif
	}
       else
#endif //DEBUG
       uart_write(antwort);
      } //Ende von Reaktion auf empfangene Zeile
    } //Ende der Hauptschlaufe
  } //Ende der Bootloader-Routine
 
#ifdef DEBUG
#ifdef TESTAUFLCD
 glcd_print("jmp 0x0000\n");
#endif
 PORTL=0x81;
 _delay_ms(2000);
#endif

 /* vor Rücksprung eventuell benutzte Hardware deaktivieren
    und Interrupts global deaktivieren, da kein "echter" Reset erfolgt */
 cli();
 uart_off(); //Serielle Schnittstelle wieder abschalten
#ifdef TESTLEDS
 PORTL=0;
 DDRL=0;
#endif
#ifdef TESTAUFLCD
 glcd_clear();
 glcd_licht(false);
 glcd_off();
#endif

 /* Interrupt Vektoren wieder gerade biegen */
 temp = MCUCR;
 MCUCR = temp | (1<<IVCE);
 MCUCR = temp & ~(1<<IVSEL);

 /* Rücksprung zur Adresse 0x0000 */
 EIND = 0; //nur auf Controllern mit mehr als 64KB Flash noetig
 start();
 return 0; //wird nie erreicht
}
