/* cbootloader.c           letzte Aenderung: 8.10.2012
Einfacher Bootloader fuer Atmega2560

Autor: Rolf Pfister
Copyright: Opensource Freeware
*/
#define VERSION "0.4"
/*
29.9.2012    Erstellung
7.10.2012    Version 0.2 (bisher nur EEPROM brennen und lesen)
             Version 0.3 jetzt auch mit Flash lesen, aber noch nicht brennen

Anwendung:
----------
> Programm (z.B. Hexprog) am Computer starten
> Reset am Atmega2560 druecken
Abwarten bis Uebertragung fertig.

Ablauf des Datenaustauschs:
Bei einem Reset am Atmega2560 sendet der Bootloader ein Startzeichen ("A\n") und
wartet eine gewisse Zeit (35ms) auf Antwort.
Hexprog antwortet dann mit einem Schluesselwort ("StartBootloader")
Bootloader sendet dann "ok", oder bei falschem Schluesselwort "error".
Wenn keine Antwort kommt oder falsches Schluesselwort wird das Anwenderprogramm
bei 0x0000 gestartet.
Sonst beginnt die Bootloader-Routine
Hexprog sendet dann ein Kommando: "check", "readeprom", "readflash", "writeeprom" oder "writeflash".

Bei den read-Kommandos wird dann vom Bootloader eine Zeile im hex-Format gesendet,
ein "ok" von Hexprog abgewartet, naechste Zeile... bis ganzer Speicher ausgelesen.
Dann sendet der Bootloader ein "ok" um das Ende der Uebertragung anzuzeigen.

Bei den write-Kommandos sendet Hexprog eine Zeile im hex-Format, wartet "ok" vom
Bootloader ab, naechste Zeile... bis alle Daten uebertragen.
Dann sendet Hexprog "ok" um Ende der Uebertragung anzuzeigen.

Bei "check" wird vom Bootloader als Antwort jeweils die empfangene Zeile wieder zurueckgeschickt.
Bei "reset" wird das Anwenderprogramm bei 0x0000 gestartet.
*/

#define ATMEGA2560
#define F_CPU 16000000UL
#define BAUD  500000UL  //Baudrate
//#define BAUD 9600UL   //falls hohe Baudrate nicht geht
#define MAXMAXFLASHADR 0x40000L  //Groesse des Flashs inklusive Bootloader
#define MAXFLASHADRESS 0x3E000L  //Groesse des Flashs ohne Bootloader

#define WARTEZEIT1 100  //Anzahl MilliSekunden zum auf Schluesselwort warten
//#define WARTEZEIT1 2000 //test
#define WARTEZEIT2 3000 //Anzahl MilliSekunden wenn spaeter auf eine Antwort gewartet wird

//#define INTERRUPTS //falls Interrupts im Bootloader gebraucht werden. Bisher nicht der Fall

//#define TESTAUFLCD  //Testausdrucke auf dem LCD beim MK3
#define TESTLEDS    //Tests auf den LEDs am PORTL
//#define DEBUG       //Zusaetzliche Tests zur Fehlersuche

#include <avr/io.h>
#ifdef INTERRUPTS
#include <avr/interrupt.h>
#endif
#include <avr/boot.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <string.h>
#include "uart.c"

#define ulong uint32_t

#ifdef TESTAUFLCD
#define LCD_MK3
#include "lcd_treiber.c"
#endif

#define MAXZ 80 //Maximale Zeilenlaenge
#define MAXD 64 //Maximale Anzahl Bytes im Datenpuffer
#define ERR_TIMEOUT 1

int8_t uart_zeile_empfangen(char *zeile,long timeout)
{
 char c;
 uint8_t i=0;
#if(F_CPU==16000000UL)
 timeout<<=10; //Bei 16MHz mit ca. 1000 multiplizieren
#else
 timeout<<=9; //Bei 8MHz weniger stark multiplizieren
#endif
 while(timeout>0)
  {if(uart_getcheck())
    {
     //c=uart_getc();
     c=UDR3;
     if(c=='\n' || i==MAXZ-1) {zeile[i]=0; return 0;}
     else if(c>=' ') zeile[i++]=c;
     else if(c=='\t') zeile[i++]=' '; //Tabulator durch Blank ersetzen, andere Steuerzeichen (z.B. '\r') ignorieren
    }
   else  --timeout;
  }
 zeile[i]=0;
 return ERR_TIMEOUT;
}

void uart_off()
{
 UCSR3B = 0;           //Auf Defaultwerte nach Datenblatt zuruecksetzen
 UCSR3C = (3<<UCSZ30); //Auf Defaultwerte nach Datenblatt zuruecksetzen
 UBRR3 = 0;            //Auf Defaultwerte nach Datenblatt zuruecksetzen
}

char hexdigit(char c)
{
 c &= 0x0F;
 if(c<0x0A) return c+'0';
 return c-0x0A+'A';
}

uint32_t char2long(uint8_t *s)
{
 //uint32_t n = ((long)s[0]<<24)+((long)s[1]<<16)+((long)s[2]<<8)+s[3];
 uint32_t n=0;
 uint8_t k;
 for(k=32;k!=0;)
   n += ((long)(*s++)) << (k-=8);
 return n;
}

// Vordeklarationen:
int8_t checksumtest(const char *s);
int8_t hexformat_kopf(char *zeile,int8_t reclen,uint16_t adresse,int8_t rectyp);
void hexformat_data(char *zeile,char c);
void hexformat_abschluss(char *zeile,int8_t checksum);
uint8_t hexformat_auspacken(char *zeile,uint16_t *adresse,uint8_t *datenpuffer,uint8_t *rectyp);
uint8_t flash_lesen(uint32_t adr);
void flash_write(uint32_t adr,uint8_t c);

int main()
{
 char zeile[MAXZ],antwort[MAXZ];
 uint8_t datenpuffer[MAXD];
 int8_t err;
 void (*start)( void ) = 0x0000;  /* Funktionspointer auf 0x0000 */
#ifdef INTERRUPTS
 uint8_t temp, n;
#endif

#ifdef TESTAUFLCD
 uint8_t nzeilen=0;
#define maxzeilen 5
 glcd_init();
 glcd_clear();
 glcd_licht(true);
#endif
#ifdef TESTLEDS
 DDRL = 0xFF;
 PORTL = 0x05; //Beim Starten ein paar LEDs einschalten
#endif

 uart_init(); //Serielle Schnittstelle initialisieren

#ifdef INTERRUPTS
 /* Interrupt Vektoren verbiegen: */
 temp = MCUCR;
 MCUCR = temp | (1<<IVCE);
 MCUCR = temp | (1<<IVSEL);
 sei();       //Interrupts einschalten
#endif

#ifdef TESTAUFLCD
 glcd_print("cbootloader Vers ");
 glcd_print(VERSION);
 glcd_print("\n");
#endif
#ifdef TESTLEDS
 PORTL = 0; //LEDs wieder aus
#endif

 /* Startbyte senden und auf Antwort mit Schluesselwort warten: */
#ifdef TESTAUFLCD
 glcd_print("gesendet:\"A\\n\"\n");
#endif
 uart_write("A\n");
 err=uart_zeile_empfangen(zeile,WARTEZEIT1);
 if(err!=0 || strcmp(zeile,"StartBootloader")!=0)
  {
#ifdef DEBUG
   if(err==ERR_TIMEOUT) strcpy(antwort,"err-timeout1\n");
   else                 strcpy(antwort,"err-StartBootloader\n");
#else
   if(err==ERR_TIMEOUT) antwort[0]=0; //Timeout ist Normalfall wenn der Bootloader nicht verwendet wird
   else                 strcpy(antwort,"error\n");
#endif
#ifdef TESTAUFLCD
   glcd_print("empfangen:'"); glcd_print(zeile); glcd_print("'\n"); _delay_ms(2000);
#endif
   if(antwort[0]!=0) uart_write(antwort);
  }
 else //Bootloader-Routine nur machen wenn korrekt geantwortet wurde
  {
   uint8_t modus=0;
#define MODUS_CHECK 1
#define MODUS_WRITEFLASH 2
#define MODUS_WRITEEPROM 3
#define MODUS_READFLASH 4
#define MODUS_READEPROM 5
   uint16_t adresse=0; //Adresse fuer EEPROM
   uint32_t longadress=0,nextadress=0,endlongadr=MAXFLASHADRESS; //Adressen fuer Flash
   uint32_t longoffset=0; //Offsetadresse zum ins Flash schreiben
   uint8_t rectyp;
#ifdef TESTAUFLCD
   glcd_print("Empfangen:'"); glcd_print(zeile); glcd_print("'\n"); _delay_ms(2000);
#endif
   uart_write("ok\n");

   while(1) //Hauptschlaufe, wird nur mit break verlassen
    {
     err=uart_zeile_empfangen(zeile,WARTEZEIT2);
     if(err==ERR_TIMEOUT)
      {
#ifdef TESTLEDS
       PORTL = 3; _delay_ms(100); PORTL = 0; //test: 2 LEDs kurz blinken lassen
#endif
#ifdef TESTAUFLCD
       if(++nzeilen>maxzeilen)  {glcd_clear(); nzeilen=0;}
       glcd_print("TIMEOUT\n");
#endif
       uart_write("err-timeout2\n");
      }
     else //andernfalls Reaktion auf die empfangene Zeile 
      {
       uint8_t n,i;
#ifdef TESTAUFLCD
       if(++nzeilen>maxzeilen)  {glcd_clear(); nzeilen=0;}
       glcd_print(zeile); glcd_print("\n"); //Empfangene Zeile auf dem LCD anzeigen
#endif
       if(strcmp(zeile,"reset")==0) break; //bei "reset" Hauptschlaufe verlassen und Anwenderprogramm starten
       antwort[0]=0; //noch keine Antwort senden
       if(modus==0) //Anfangs auf Kommandos zum Modus setzen warten
	{if(strcmp(zeile,"writeflash")==0) {modus=MODUS_WRITEFLASH; strcpy(antwort,"ok\n");}
	 else if(strcmp(zeile,"writeeprom")==0) {modus=MODUS_WRITEEPROM; strcpy(antwort,"ok\n");}
	 else if(strcmp(zeile,"readflash")==0)  {modus=MODUS_READFLASH; endlongadr=MAXFLASHADRESS-1;}
	 else if(strcmp(zeile,"readflash2")==0) {modus=MODUS_READFLASH; endlongadr=MAXMAXFLASHADR-1;}
	 else if(strcmp(zeile,"readeprom")==0)  {modus=MODUS_READEPROM;}
	 else if(strcmp(zeile,"check")==0)      {modus=MODUS_CHECK; strcpy(antwort,"check\n");}
	 else strcpy(antwort,"err-command\n");
	}
       else if(modus==MODUS_CHECK)
	{
	 strcpy(antwort,zeile); antwort[n=strlen(zeile)]='\n'; antwort[n+1]=0;
	}
       else if(modus==MODUS_WRITEFLASH)
	{
	 if(strcmp(zeile,"ok")==0) break; //Ende der Uebertragung --> Hauptschlaufe verlassen
	 else if(zeile[0]!=':') strcpy(antwort,"err-colon\n");
	 else if(checksumtest(&zeile[1])!=0) strcpy(antwort,"err-checksum\n");
	 else
	  {
#ifdef TESTAUFLCD
	   glcd_print("Daten werden ins Flash gebrannt\n");
#endif
	   n=hexformat_auspacken(zeile,&adresse,datenpuffer,&rectyp);
	   if(rectyp==0)
	    {
	     longadress = longoffset+adresse;
	     for(i=0;i<n;i++)
	       flash_write(longadress+i,datenpuffer[i]);
	     strcpy(antwort,"ok\n");
	    }
	   else if(rectyp==1) //End of File Record
	    {
	     flash_write(0L,0); //eventuell ausstehende Page noch speichern
	     strcpy(antwort,"ok\n");
	    }
	   else if(rectyp==2) //Extended Segment Address Record
	    {
	     longoffset = char2long(datenpuffer)<<4;
	     strcpy(antwort,"ok\n");
	    }
	   else //andere Rectypen noch nicht ausgewertet
	    {
	     strcpy(antwort,"err-unknownrectyp\n");
	    }
	  }
	}
       else if(modus==MODUS_WRITEEPROM)
	{
	 if(strcmp(zeile,"ok")==0) break; //Ende der Uebertragung --> Hauptschlaufe verlassen
	 else if(zeile[0]!=':') strcpy(antwort,"err-colon\n");
	 else if(checksumtest(&zeile[1])!=0) strcpy(antwort,"err-checksum\n");
	 else
	  {
	   uint8_t n,i;
#ifdef TESTAUFLCD
	   glcd_print("Daten werden ins EEPROM gebrannt\n");
#endif
	   n=hexformat_auspacken(zeile,&adresse,datenpuffer,&rectyp);
	   if(rectyp==0)
	    for(i=0;i<n;i++)
	      eeprom_write_byte((uint8_t*)(adresse+i),datenpuffer[i]);
	   //else if(rectyp==1) {} //andere Rectypen fuer EEPROM nicht ausgewertet, da nicht sinnvoll
	   strcpy(antwort,"ok\n");
	  }
	}
       if(modus==MODUS_READFLASH)  //kein else weil sofort aufs Kommando "readflash" die erste Zeile gesendet werden soll
	{
	 int8_t i,j,checksum;
	 uint8_t c;
	 static int8_t offsetflag=0;
	 if(strcmp(zeile,"ok")==0) //Wenn letztes Senden erfolgreich war
	  {if(offsetflag!=0)       //falls Offsetadresse gesendet wurde, dann Flag ruecksetzen um
	    {if(offsetflag==2) longadress=nextadress; //Daten von neuer Adresse zu senden
	     offsetflag=0;                            //oder Daten von selber Adresse zu senden
	    }
	   else                    //sonst Adresse erhoehen
	    {longadress += 16;
	     if((longadress&0xFFFF)==0 && longadress<endlongadr)
	          offsetflag=1;   //Bei neuem 64KB Block Markierung dass Offsetadresse gesendet werden soll
	     else offsetflag=0;
	    }
	  }
	 else if(strncmp(zeile,"readf",5)==0) //Beim ersten Mal
	  {longadress=0;                      //bei Adresse 0 beginnen
	   while(flash_lesen(longadress)==0xFF) longadress++; //erste nicht geloeschte Stelle suchen
	   if(endlongadr<=longadress) endlongadr=MAXMAXFLASHADR-1;
	   while(flash_lesen(endlongadr)==0xFF) endlongadr--; //letzte nicht geloeschte Stelle suchen
	   endlongadr++; //Endadress eins hoeher setzen
	   if(longadress>=0x10000L) {nextadress=longadress; offsetflag=2; longadress &= 0xF0000L;}
	  }
	 else           //Im Fehlerfall
	  {             //Daten von selber Adresse nochmals senden, oder neuer Offset nochmals senden
	  }
	 if(longadress>=endlongadr+16)
	  {uart_write("ok\n"); break;} //wenn fertig: "ok" senden und Hauptschlaufe verlassen
	 else if(longadress>=endlongadr)
	   strcpy(antwort,":00000001FF\n"); //wenn Endadresse erreicht: Abschlusszeile senden
	 else
	  {
	   if(offsetflag!=0)
	    { //neue Offsetadresse
	     checksum=hexformat_kopf(antwort,2,0x0000,2); j=9; //Extended Segment Address Record (Typ 02)
	     c = (longadress>>12); //4 Bits nach rechts geschoben, und dann oberes Byte davon
	     hexformat_data(&antwort[j],c); checksum+=c; j+=2;
	     c = 0;                //unteres Byte immer gleich 0
	     hexformat_data(&antwort[j],c); j+=2; //checksum+=c;
	    }
	   else
	    {
	     checksum=hexformat_kopf(antwort,16,longadress&0xFFFF,0); j=9; //Kopfdaten fuers Intel-Hex-Format ergibt 9 Bytes
	     for(i=0;i<16;i++)
	      {
	       c = __ELPM_classic__(longadress+i);
	       hexformat_data(&antwort[j],c); checksum+=c; j+=2;
	      }
	    }
	   hexformat_abschluss(&antwort[j],checksum);
	  }
	}
       else if(modus==MODUS_READEPROM)
	{
	 int8_t i,j,checksum;
	 if(strcmp(zeile,"ok")==0) //Wenn letzte 16 Zeichen erfolgreich empfangen wurden
	   adresse += 16;          //dann Adresse erhoehen. (Sonst Daten von selber Adresse nochmals senden)
	 if(adresse==0x1000)
	   strcpy(antwort,":00000001FF\n"); //wenn Endadresse erreicht: Abschlusszeile senden
	 else if(adresse>0x1000)
	  {uart_write("ok\n"); break;} //wenn fertig: "ok" senden und Hauptschlaufe verlassen
	 else
	  {
	   checksum=hexformat_kopf(antwort,16,adresse,0); j=9; //Kopfdaten fuers Intel-Hex-Format ergibt 9 Bytes
	   for(i=0;i<16;i++)
	    {
	     char c = eeprom_read_byte((uint8_t*)(adresse+i));
	     hexformat_data(&antwort[j],c); checksum+=c; j+=2;
	    }
	   hexformat_abschluss(&antwort[j],checksum);
	  }
	}
#ifdef DEBUG
       if(antwort[0]==0) //wenn alles korrekt programmiert sollte das nicht passieren
	{strcpy(antwort,"err-unknown\n");
#ifdef TESTLEDS
	 PORTL = 7; _delay_ms(100); PORTL = 0; //test: 3 LEDs kurz blinken lassen
#endif
#ifdef TESTAUFLCD
	 if(++nzeilen>maxzeilen)  {glcd_clear(); nzeilen=0;}
	 glcd_print(antwort);
#endif
	}
       else
#endif //DEBUG
       uart_write(antwort);
      } //Ende von Reaktion auf empfangene Zeile
    } //Ende der Hauptschlaufe
   _delay_ms(100);
  } //Ende der Bootloader-Routine
 
#ifdef DEBUG
#ifdef TESTAUFLCD
 glcd_print("jmp 0x0000\n");
#endif
 PORTL=0x81;
 _delay_ms(2000);
#endif

 /* vor Rücksprung eventuell benutzte Hardware deaktivieren
    und Interrupts global deaktivieren, da kein "echter" Reset erfolgt */
#ifdef INTERRUPTS
 cli();
 /* Interrupt Vektoren wieder gerade biegen */
 temp = MCUCR;
 MCUCR = temp | (1<<IVCE);
 MCUCR = temp & ~(1<<IVSEL);
#endif

 uart_off(); //Serielle Schnittstelle wieder abschalten

#ifdef TESTLEDS
 PORTL=0;
 DDRL=0;
#endif
#ifdef TESTAUFLCD
 glcd_clear();
 glcd_licht(false);
 glcd_off();
#endif

 /* Rücksprung zur Adresse 0x0000 */
 EIND = 0; //nur auf Controllern mit mehr als 64KB Flash noetig
 start();
 return 0; //wird nie erreicht
}

// Routinen fuers Intel-Hex-Format
// Beschreibung zum Format siehe  de.wikipedia.org/wiki/Intel_HEX
int8_t hexformat_kopf(char *zeile,int8_t reclen,uint16_t adresse,int8_t rectyp)
{
 int8_t checksum;
 *zeile++ = ':'; //Startzeichen fuers Intel-hex-Format
 *zeile++ = hexdigit(reclen>>4);
 *zeile++ = hexdigit(reclen);
 checksum = reclen;
 *zeile++ = hexdigit(adresse>>12);
 *zeile++ = hexdigit(adresse>>8);
 checksum += (adresse>>8);
 *zeile++ = hexdigit(adresse>>4);
 *zeile++ = hexdigit(adresse);
 checksum += (adresse&0xFF);
 *zeile++ = hexdigit(rectyp>>4);
 *zeile++ = hexdigit(rectyp);
 checksum += rectyp;
 return checksum;
}

void hexformat_data(char *zeile,char c)
{
 *zeile++ = hexdigit(c>>4);
 *zeile++ = hexdigit(c);
}

void hexformat_abschluss(char *zeile,int8_t checksum)
{
 checksum = -checksum;
 *zeile++ = hexdigit(checksum>>4);
 *zeile++ = hexdigit(checksum);
 *zeile++ = '\n';
 *zeile++ = 0;
}

char ascitodigit(char c)
{		/* wandelt ein ASCI-Zeichen in eine 1-stellige Hex-Zahl */
 if(c>='0' && c<='9') return c-'0';
 if(c>='A' && c<='F') return c-'A'+10;
 if(c>='a' && c<='f') return c-'a'+10;
 return 0;
 /*
 c -= '0';
 if(c>9) c -= ('A'-'9'-1);
 if(c>0x0F) c -= ('a'-'A');
 if(c<0 || c>0x0F) c=0;
 return c;
 */
}

int8_t checksumtest(const char *s)
{
 int8_t pruefsumme=0,c1,c2;
 while((c1= *s++)!=0 && (c2= *s++)!=0)
  {
   pruefsumme += (ascitodigit(c1)<<4)+ascitodigit(c2);
  }
 return pruefsumme;
}

uint8_t hexformat_auspacken(char *zeile,uint16_t *adresse,uint8_t *datenpuffer,uint8_t *rectyp)
{
 uint8_t reclen,i,j;
 uint16_t c1,c2;
 reclen = (ascitodigit(zeile[1])<<4)+ascitodigit(zeile[2]);
 *rectyp = (ascitodigit(zeile[7])<<4)+ascitodigit(zeile[8]);
 c1 = (ascitodigit(zeile[3])<<4)+ascitodigit(zeile[4]);
 c2 = (ascitodigit(zeile[5])<<4)+ascitodigit(zeile[6]);
 *adresse = (c1<<8)+c2;
 for(i=0,j=9;i<reclen;i++,j+=2)
  {
   *datenpuffer++ = (ascitodigit(zeile[j])<<4)+ascitodigit(zeile[j+1]);
  }
 return reclen;
}

uint8_t flash_lesen(uint32_t adr)
{
 return __ELPM_classic__(adr);
}

/*********************** Diese Routinen eventuell in Assembler schreiben ***********************/
//#define SPM_PAGESIZEB (SPM_PAGESIZE<<1)  //muesste beim Atmega2560 gleich 512 sein, oder doch 256 ?
#define SPM_PAGESIZEB (SPM_PAGESIZE)  //oder ist in boot.h schon Byteadresse gesetzt?
#define SPM_PAGEMASK (SPM_PAGESIZEB-1L)

void boot_program_page(uint32_t page, uint8_t *buf)  //von www.mikrocontroller.net kopiert
{
    uint16_t i;
#ifdef INTERRUPTS
    uint8_t sreg = SREG;
    cli();     // Disable interrupts.
#endif
    eeprom_busy_wait();
    boot_page_erase(page);
    boot_spm_busy_wait();     // Wait until the memory is erased.
    for (i=0; i<SPM_PAGESIZEB; i+=2)
    {
        uint16_t w = *buf++;   // Set up little-endian word.
        w += (*buf++) << 8;
        boot_page_fill(page+i,w);
    }
    boot_page_write(page);     // Store buffer in flash page.
    boot_spm_busy_wait();      // Wait until the memory is written.
 
    // Reenable RWW-section again. We need this if we want to jump back
    // to the application after bootloading.
    boot_rww_enable();
 
#ifdef INTERRUPTS
    SREG = sreg;    // Re-enable interrupts (if they were ever enabled).
#endif
}

void flash_write(uint32_t adr,uint8_t c)
{
 static uint8_t puffer[SPM_PAGESIZEB];
 static uint16_t j=0;
 static uint32_t page=0;
#ifdef TESTLEDS
 PORTL=0xF; //4 LEDs ein
#endif
 if(j==0)
  {
   page = adr & ~SPM_PAGEMASK;
  }
 if(adr!=page+j)
  {
   if(adr<page+j || adr-page>=SPM_PAGESIZEB)
    {
     while(j<SPM_PAGESIZEB) puffer[j++]=0xFF; //restliche page fuellen
     boot_program_page(page,puffer);          //und page schreiben
     j=0;
     page = adr & ~SPM_PAGEMASK;
    }
   while(adr!=page+j) puffer[j++]=0xFF; //Luecke in der page fuellen
  }
 puffer[j++]=c;
 if(j==SPM_PAGESIZEB)
    {
     boot_program_page(page,puffer);
     j=0;
    }
#ifdef TESTLEDS
 PORTL=0; //alle aus
#endif
}
